import numpy as np
from numpy import array
import numpy.testing as npt

from limix_inference.cov import LinearCov
from optimix import check_grad

from optimix.testing import Assertion

def test_linearcov_optimix():
    item0 = array([-1.5, 1.0])
    item1 = array([+1.24, 1.0])
    a = Assertion(lambda: LinearCov(), item0, item1, 0.0, logscale=0.0)
    a.assert_layout()
    a.assert_gradient()

def test_value():
    random = np.random.RandomState(0)
    cov = LinearCov()

    x0 = random.randn(10)
    x1 = random.randn(10)

    npt.assert_almost_equal(cov.value(x0, x1), np.dot(x0, x1))
    cov.scale = 2.
    npt.assert_almost_equal(cov.value(x0, x1), 2 * np.dot(x0, x1))


def test_gradient():
    random = np.random.RandomState(0)
    cov = LinearCov()
    cov.scale = 2.

    x0 = random.randn(10)
    x1 = random.randn(10)

    def func(x):
        cov.scale = np.exp(x[0])
        return cov.value(x0, x1)

    def grad(x):
        cov.scale = np.exp(x[0])
        return [cov.derivative_logscale(x0, x1)]

    npt.assert_almost_equal(check_grad(func, grad, [2.0]), 0, decimal=6)

if __name__ == '__main__':
    __import__('pytest').main([__file__, '-s'])



from __future__ import division

from numpy import array, dot, ones, pi, sqrt, zeros
from numpy.random import RandomState
from numpy.testing import assert_almost_equal

from numpy_sugar.linalg import economic_qs_linear

from limix_inference.glmm import ExpFamEP
from limix_inference.lik import BinomialProdLik
from limix_inference.link import LogitLink

def test_binomial_lml():
    n = 3
    M = ones((n, 1)) * 1.
    G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
    (Q, S0) = economic_qs_linear(G)
    nsuccesses = array([1., 0., 1.])
    ntrials = array([1., 1., 1.])
    lik = BinomialProdLik(ntrials, LogitLink())
    lik.nsuccesses = nsuccesses
    ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1)
    ep.beta = array([1.])
    assert_almost_equal(ep.beta, array([1.]))
    ep.v = 1.
    ep.delta = 0
    assert_almost_equal(ep.lml(), -2.3202659215368935)


def test_binomial_gradient_over_v():
    n = 3
    M = ones((n, 1)) * 1.
    G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
    (Q, S0) = economic_qs_linear(G)
    nsuccesses = array([1., 0., 1.])
    ntrials = array([1., 1., 1.])
    lik = BinomialProdLik(ntrials, LogitLink())
    lik.nsuccesses = nsuccesses
    ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1)
    ep.beta = array([1.])
    assert_almost_equal(ep.beta, array([1.]))
    ep.v = 1.
    ep.delta = 0.

    analytical_gradient = ep._gradient_over_v()

    lml0 = ep.lml()
    step = 1e-5
    ep.v = ep.v + step
    lml1 = ep.lml()

    empirical_gradient = (lml1 - lml0) / step

    assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)

    ep.v = 0.5
    ep.delta = 0.0

    analytical_gradient = ep._gradient_over_v()

    lml0 = ep.lml()
    step = 1e-5
    ep.v = ep.v + step
    lml1 = ep.lml()

    empirical_gradient = (lml1 - lml0) / step

    assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)


def test_binomial_gradient_over_delta():
    n = 3
    M = ones((n, 1)) * 1.
    G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
    (Q, S0) = economic_qs_linear(G)
    nsuccesses = array([1., 0., 1.])
    ntrials = array([1., 1., 1.])
    lik = BinomialProdLik(ntrials, LogitLink())
    lik.nsuccesses = nsuccesses
    ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1.0)
    ep.beta = array([1.])
    assert_almost_equal(ep.beta, array([1.]))
    ep.v = 1.
    ep.delta = 0.5

    analytical_gradient = ep._gradient_over_delta()

    lml0 = ep.lml()
    step = 1e-5
    ep.delta = ep.delta + step
    lml1 = ep.lml()

    empirical_gradient = (lml1 - lml0) / step

    assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)


def test_binomial_gradient_over_both():
    n = 3
    M = ones((n, 1)) * 1.
    G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
    (Q, S0) = economic_qs_linear(G)
    nsuccesses = array([1., 0., 1.])
    ntrials = array([1., 1., 1.])
    lik = BinomialProdLik(ntrials, LogitLink())
    lik.nsuccesses = nsuccesses
    ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1.0)
    ep.beta = array([1.])
    assert_almost_equal(ep.beta, array([1.]))
    ep.v = 1.5
    ep.delta = 0.3

    analytical_gradient = ep._gradient_over_both()[0]

    lml0 = ep.lml()
    step = 1e-5
    ep.v = ep.v + step
    lml1 = ep.lml()

    empirical_gradient = (lml1 - lml0) / step

    assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)

    ep.v = 1.5
    ep.delta = 0.3

    analytical_gradient = ep._gradient_over_both()[1]

    lml0 = ep.lml()
    step = 1e-5
    ep.delta = ep.delta + step
    lml1 = ep.lml()

    empirical_gradient = (lml1 - lml0) / step

    assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)


def test_binomial_optimize():
    random = RandomState(139)
    nsamples = 30
    nfeatures = 31

    G = random.randn(nsamples, nfeatures) / sqrt(nfeatures)

    u = random.randn(nfeatures)

    z = 0.1 + 2 * dot(G, u) + random.randn(nsamples)

    ntrials = random.randint(10, 500, size=nsamples)

    y = zeros(nsamples)
    for i in range(len(ntrials)):
        y[i] = sum(z[i] + random.logistic(
            scale=pi / sqrt(3), size=ntrials[i]) > 0)
    (Q, S0) = economic_qs_linear(G)

    M = ones((nsamples, 1))
    lik = BinomialProdLik(ntrials, LogitLink())
    lik.nsuccesses = y
    ep = ExpFamEP(lik, M, Q[0], Q[1], S0)
    ep.optimize()

    assert_almost_equal(ep.lml(), -144.2381842202486, decimal=3)


if __name__ == '__main__':
    __import__('pytest').main([__file__, '-s'])



"""
*******************************
Generalized Linear Mixed Models
*******************************

Example
^^^^^^^

.. doctest::

    >>> from limix_inference.random import bernoulli_sample
    >>> from limix_inference.glmm import ExpFamEP
    >>> from limix_inference.lik import BernoulliProdLik
    >>> from limix_inference.link import LogLink
    >>> from numpy_sugar.linalg import economic_qs_linear
    >>> from numpy.random import RandomState
    >>> offset = 5
    >>> G = [[1, -1], [2, 1], [1, 3], [2, 1]]
    >>> (Q0, Q1), S0 = economic_qs_linear(G)
    >>> y = bernoulli_sample(offset, G, random_state=RandomState(0))
    >>> covariates = [[1.], [0.6], [1.], [-0.6]]
    >>> lik = BernoulliProdLik(LogLink)
    >>> lik.outcome = y
    >>> glmm = ExpFamEP(lik, covariates, Q0, Q1, S0)
    >>> glmm.optimize()
    >>> glmm.lml()
    -1.2067335984780192

Expectation propagation
^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: ExpFamEP
    :members: covariates_variance, genetic_variance, environmental_variance,
              heritability, K, m, beta, v, delta, lml, optimize, fixed_ep
"""

from ._ep import ExpFamEP



from __future__ import division

from numpy.testing import assert_allclose

from numpy import exp
from numpy import arange

from optimix import check_grad

from limix_inference.cov import EyeCov

from optimix.testing import Assertion

def test_eyecov_optimix():
    item0 = 0
    item1 = 1
    a = Assertion(lambda: EyeCov(), item0, item1, 0.0, logscale=0.0)
    a.assert_layout()

def test_eye_value():
    cov = EyeCov()
    cov.scale = 2.1
    assert_allclose(cov.value(0, 0), 2.1)


def test_eye_gradient_1():
    cov = EyeCov()
    cov.scale = 2.1
    a = arange(1)
    cov.set_data((a, a))

    def func(x):
        cov.scale = exp(x[0])
        return cov.feed().value()

    def grad(x):
        cov.scale = exp(x[0])
        return cov.feed().gradient()

    assert_allclose(check_grad(func, grad, [0.1]), 0, atol=1e-7)


def test_eye_gradient_2():
    cov = EyeCov()
    cov.scale = 2.1
    a = arange(5)
    cov.set_data((a, a))

    def func(x):
        cov.scale = exp(x[0])
        return cov.feed().value()

    def grad(x):
        cov.scale = exp(x[0])
        return cov.feed().gradient()

    assert_allclose(check_grad(func, grad, [0.1]), 0, atol=1e-7)

if __name__ == '__main__':
    __import__('pytest').main([__file__, '-s'])



class GP2KronSum(object):
    """
    Gaussian Process with a 2kronSum Covariance and a mean that is a sum of Kronecker products:
        vec(Y) ~ N( vec( \sum_i F_i B_i A_i), Cg \kron R + Cn \kron I )
    Notation:
        N = number of samples
        P = number of traits
        Y = [N, P] phenotype matrix
        F_i = sample fixed effect design for term i
        A_i = trait fixed effect design for term i
        B_i = effect sizes of fixed effect term i
        Cg = column covariance matrix for signal term respectively
        Cn = column covariance matrix for noise term respectively
        R = row covariance matrix for signal term respectively
    """

    def __init__(self, Y, Cg, Cn, F=None, A=None, R=None):
        """
        Args:
            Y:      [N, P] phenotype matrix
            F:      list of sample fixed effect designs.
                    Each term must have first dimension N
            A:      list of trait fixed effect design.
                    Each term must have second dimension P
            Cg:     Limix covariance matrix for Cg (dimension P)
            Cn:     Limix covariance matrix for Cn (dimension P)
            R:      [N, N] numpy semidemidefinite covariance matrix for R.
                    In alternative to R, S_R and U_R can be specified.
            S_R:    N vector of eigenvalues of R
            U_R:    [N, N] eigenvector matrix of R
        """
        assert_type(Y, NP.ndarray, 'Y')
        assert_subtype(Cg, Covariance, 'Cg')
        assert_subtype(Cn, Covariance, 'Cn')

        covar = Cov2KronSum(Cg=Cg, Cn=Cn, R=R, S_R=S_R, U_R=U_R)
        mean = MeanKronSum(Y=Y, F=F, A=A)

        GP.__init__(self, covar=covar, mean=mean)

    # def _observe(self):
    #     self.covar.register(self.row_cov_has_changed, 'row_cov')
    #     self.covar.register(self.col_cov_has_changed, 'col_cov')
    #     self.mean.register(self.pheno_has_changed, 'pheno')
    #     self.mean.register(self.designs_have_changed, 'designs')
    #
    # def col_cov_has_changed(self):
    #     self.clear_cache('col_cov')
    #     self.clear_all()
    #
    # def row_cov_has_changed(self):
    #     self.clear_cache('row_cov')
    #     self.clear_all()
    #
    # def pheno_has_changed(self):
    #     self.clear_cache('pheno')
    #     self.clear_all()
    #
    # def designs_have_changed(self):
    #     self.clear_cache('designs')
    #     self.clear_all()
    #
    #
    # ######################
    # # Transformed phenotype
    # ######################
    # @cached('row_cov', 'pheno')
    # def LrY(self):
    #     return sp.dot(self.covar.Lr(), self.mean.Y)
    #
    # @cached(['row_cov', 'col_cov', 'pheno'])
    # def LrYLc(self):
    #     return sp.dot(self.LrY(), self.covar.Lc().T)
    #
    # @cached(['row_cov', 'col_cov', 'pheno'])
    # def DLrYLc(self):
    #     return self.covar.D() * self.LrYLc()
    #
    # ######################
    # # Transformed fixed effects
    # ######################
    # @cached(['row_cov', 'designs'])
    # def LrF(self):
    #     R = []
    #     for ti in range(self.mean.n_terms):
    #         R.append(sp.dot(self.covar.Lr(), self.mean.F[ti]))
    #     return R
    #
    # @cached(['col_cov', 'designs'])
    # def ALc(self):
    #     R = []
    #     for ti in range(self.mean.n_terms):
    #         R.append(sp.dot(self.mean.A[ti], self.covar.Lc().T))
    #     return R
    #
    # @cached(['row_cov', 'col_cov', 'designs'])
    # def LW(self):
    #     R = sp.zeros((self.mean.Y.size, self.mean.n_covs))
    #     istart = 0
    #     for ti in range(self.mean.n_terms):
    #         iend = istart + self.mean.F[ti].shape[1] * self.mean.A[ti].shape[0]
    #         R[:, istart:iend] = sp.kron(self.ALc()[ti].T, self.LrF()[ti])
    #         istart += iend
    #     return R
    #
    # @cached(['row_cov', 'col_cov', 'designs'])
    # def dLW(self):
    #     return self.covar.d()[:,sp.newaxis] * self.LW()
    #
    # @cached(['row_cov', 'col_cov', 'pheno'])
    # def Sr_DLrYLc_Ctilde(self, i):
    #     return self.covar.Sr_X_Ctilde(self.DLrYLc(), i)
    #
    # @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    # def Sr_vei_dLWb_Ctilde(self, i):
    #     return self.covar.Sr_X_Ctilde(self.vei_dLWb(), i)
    #
    # @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    # def vei_dLWb(self):
    #     # could be optimized but probably not worth it
    #     # as it requires a for loop
    #     r = sp.dot(self.dLW(), self.mean.b)
    #     return r.reshape(self.mean.Y.shape, order = 'F')
    #
    # ######################
    # # Areml
    # ######################
    # def Areml_K(self):
    #     return sp.dot(self.LW().T, self.dLW())
    #
    # def Areml_K_grad_i(self,i):
    #     dLWt = self.dLW().reshape((self.mean._N, self.mean._P, self.mean.n_covs), order = 'F')
    #     if i < self.covar.Cg.getNumberParams():
    #         SrdLWt = self.covar.Sr()[:, sp.newaxis, sp.newaxis] * dLWt
    #     else:
    #         SrdLWt = dLWt
    #     SrdLWtC = sp.tensordot(SrdLWt, self.covar.Ctilde(i), axes=(1, 1))
    #     SroCdLW = SrdLWtC.swapaxes(1,2).reshape((self.mean._N * self.mean._P, self.mean.n_covs), order = 'F')
    #     return -sp.dot(self.dLW().T, SroCdLW)
    #
    # ########################
    # # LML terms
    # ########################
    # @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    # def WKiy(self):
    #     R = sp.zeros((self.mean.n_covs, 1))
    #     istart = 0
    #     for ti in range(self.mean.n_terms):
    #         _dim = self.mean.F[ti].shape[1] * self.mean.A[ti].shape[0]
    #         iend = istart + _dim
    #         FLrDLrYLc = sp.dot(self.LrF()[ti].T, self.DLrYLc())
    #         R[istart:iend, 0] = sp.dot(FLrDLrYLc, self.ALc()[ti].T).reshape(_dim, order = 'F')
    #     return R
    #
    # @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    # def b(self):
    #     if self.mean.n_covs > 0:
    #         R = self.Areml.solve(self.WKiy())
    #     else:
    #         R = None
    #     return R
    #
    # @cached(['row_cov', 'col_cov', 'pheno'])
    # def yKiy(self):
    #     return (self.LrYLc()*self.DLrYLc()).sum()
    #
    # @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    # def yKiWb(self):
    #     return (self.LrYLc() * self.vei_dLWb()).sum()
    #
    # #########################
    # # Gradients
    # #########################
    # @cached(['row_cov', 'col_cov', 'pheno'])
    # def yKiy_grad_i(self,i):
    #     return -(self.DLrYLc() * self.Sr_DLrYLc_Ctilde(i)).sum()
    #
    # @cached(['row_cov', 'col_cov', 'designs', 'pheno'])
    # def yKiWb_grad_i(self,i):
    #     rv = -2*(self.DLrYLc()*self.Sr_vei_dLWb_Ctilde(i)).sum()
    #     rv+= (self.vei_dLWb()*self.Sr_vei_dLWb_Ctilde(i)).sum()
    #     return rv
    #
    # ############################
    # # Simulate
    # ############################
    # def simulate_pheno(self, Rh=None):
    #     if Rh is None:
    #         S, U = la.eigh(self.covar.R)
    #         S[S<0] = 0
    #         Rh = U*S**(0.5)
    #     Yc = sp.dot(self.mean.F[0], sp.dot(self.mean.B[0], self.mean.A[0].T))
    #     Z = sp.randn(Rh.shape[1], self.covar.Cr.X.shape[1])
    #     Yr = sp.dot(Rh, sp.dot(Z, self.covar.Cr.X.T))
    #     _S, _U = LA.eigh(self.covar.Cn.K()); _S[_S<0] = 0
    #     Cn_h = _U*_S**0.5
    #     Yn = sp.dot(sp.randn(*self.mean.Y.shape), Cn_h.T)
    #     RV = Yc+Yr+Yn
    #     return RV



from __future__ import division

from numpy import array, dot, ones, sqrt, zeros, exp
from numpy.random import RandomState
from numpy.testing import assert_almost_equal

from numpy_sugar.linalg import economic_qs_linear
from limix_inference.glmm import ExpFamEP
from limix_inference.lik import PoissonProdLik
from limix_inference.link import LogLink


def test_poisson_lml():
    n = 3
    M = ones((n, 1)) * 1.
    G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
    (Q, S0) = economic_qs_linear(G)
    noccurrences = array([1., 0., 5.])
    lik = PoissonProdLik(LogLink())
    lik.noccurrences = noccurrences
    ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1)
    ep.beta = array([1.])
    assert_almost_equal(ep.beta, array([1.]))
    ep.v = 1.
    ep.delta = 0
    assert_almost_equal(ep.lml(), -6.793765561069963)


def test_poisson_optimize():
    random = RandomState(139)
    nsamples = 30
    nfeatures = 31

    G = random.randn(nsamples, nfeatures) / sqrt(nfeatures)

    u = random.randn(nfeatures)

    z = 0.1 + 2 * dot(G, u) + random.randn(nsamples)

    y = zeros(nsamples)
    for i in range(nsamples):
        y[i] = random.poisson(lam=exp(z[i]))
    (Q0, Q1), S0 = economic_qs_linear(G)

    M = ones((nsamples, 1))
    lik = PoissonProdLik(LogLink())
    lik.noccurrences = y
    ep = ExpFamEP(lik, M, Q0, Q1, S0)
    ep.optimize()
    assert_almost_equal(ep.lml(), -77.90919831238075, decimal=2)
    assert_almost_equal(ep.beta[0], 0.314709077094, decimal=1)
    assert_almost_equal(ep.heritability, 0.797775054939, decimal=1)


if __name__ == '__main__':
    __import__('pytest').main([__file__, '-s'])



"""
*******************
Linear Mixed Models
*******************

- :class:`.FastLMM`
- :class:`.SlowLMM`

Fast implementation
^^^^^^^^^^^^^^^^^^^

.. autoclass:: FastLMM
  :members:

General implementation
^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: SlowLMM
  :members:
"""

from ._fastlmm import FastLMM
from ._slowlmm import SlowLMM



from numpy.testing import assert_allclose

from limix_inference.link import ProbitLink


def test_probit_link():
    lik = ProbitLink()
    assert_allclose(lik.value(lik.inv(3.2)), 3.2)


if __name__ == '__main__':
    __import__('pytest').main([__file__, '-s'])



from __future__ import division

from numpy import array, newaxis, zeros
from numpy.testing import assert_allclose

from limix_inference.cov import FreeFormCov
from optimix import check_grad

from optimix.testing import Assertion

# def test_freeform_optimix():
#     item0 = 0
#     item1 = 1
#     a = Assertion(lambda: FreeFormCov(2), item0, item1, 0.0)
#     a.assert_layout()
#
# def test_freeform_value():
#     cov = FreeFormCov(2)
#     assert_allclose(cov.value(0, 0), 1)
#     assert_allclose(cov.value(0, 1), 1)
#     assert_allclose(cov.value(1, 0), 1)
#     assert_allclose(cov.value(1, 1), 2)
#
#     assert_allclose(cov.value([[0], [1]], [[0], [1]]), [[1, 1], [1, 2]])
#
#
# def test_freeform_derivate():
#     cov = FreeFormCov(2)
#
#     step = 1e-7
#
#     x0 = array([1, 1, 1], float)
#     cov.set('Lu', x0)
#     y0 = cov.value(0, 0)
#
#     x1 = array([1 + step, 1, 1], float)
#     cov.set('Lu', x1)
#     y1 = cov.value(0, 0)
#
#     assert_allclose(
#         (y1 - y0) / step, cov.derivative_Lu(0, 0)[0, 0, 0], rtol=1e-5)
#
#     x0 = array([1, 1, 1], float)
#     cov.set('Lu', x0)
#     y0 = cov.value(0, 0)
#
#     x1 = array([1, 1 + step, 1], float)
#     cov.set('Lu', x1)
#     y1 = cov.value(0, 0)
#
#     assert_allclose(
#         (y1 - y0) / step, cov.derivative_Lu(0, 0)[0, 0, 1], rtol=1e-5)
#
#     x0 = array([1, 1, 1], float)
#     cov.set('Lu', x0)
#     y0 = cov.value(0, 0)
#
#     x1 = array([1, 1, 1 + step], float)
#     cov.set('Lu', x1)
#     y1 = cov.value(0, 0)
#
#     assert_allclose(
#         (y1 - y0) / step, cov.derivative_Lu(0, 0)[0, 0, 2], rtol=1e-5)
#
#
# def test_freeform_gradient():
#     cov = FreeFormCov(2)
#     idx = array([[0], [1]], int)
#     cov.set_data((idx, idx))
#
#     def func(x):
#         Lu = cov.Lu
#         Lu[0] = x[0]
#         cov.Lu = Lu
#         return cov.feed().value()
#
#     def grad(x):
#         Lu = cov.Lu
#         Lu[0] = x[0]
#         cov.Lu = Lu
#         return [cov.feed().gradient()[0][:, :, 0]]
#
#     assert_allclose(check_grad(func, grad, [1.0], step=1e-5), 0, atol=1e-5)

# def test_crap():
#     item0 = 0
#     item1 = 1
#     a = Assertion(lambda: FreeFormCov(2), item0, item1, 0.0, Lu=zeros(3))
#     a.assert_layout()

if __name__ == '__main__':
    __import__('pytest').main([__file__, '-s'])



from __future__ import division

import numpy as np
from numpy import ones
from numpy import arange
from numpy.testing import assert_allclose

from numpy_sugar.linalg import economic_qs_linear

from limix_inference.lmm import FastLMM
from limix_inference.lik import DeltaProdLik
from limix_inference.cov import LinearCov
from limix_inference.cov import EyeCov
from limix_inference.cov import SumCov
from limix_inference.mean import OffsetMean
from limix_inference.random import GLMMSampler


def test_learn():
    random = np.random.RandomState(9458)
    N = 500
    X = random.randn(N, N + 1)
    X -= X.mean(0)
    X /= X.std(0)
    X /= np.sqrt(X.shape[1])
    offset = 1.0

    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(N, purpose='sample')

    cov_left = LinearCov()
    cov_left.scale = 1.5
    cov_left.set_data((X, X), purpose='sample')

    cov_right = EyeCov()
    cov_right.scale = 1.5
    cov_right.set_data((arange(N), arange(N)), purpose='sample')

    cov = SumCov([cov_left, cov_right])

    lik = DeltaProdLik()

    y = GLMMSampler(lik, mean, cov).sample(random)

    (Q0, Q1), S0 = economic_qs_linear(X)

    flmm = FastLMM(y, Q0, Q1, S0, covariates=ones((N, 1)))

    flmm.learn()

    assert_allclose(flmm.beta[0], 0.8997652129631661, rtol=1e-5)
    assert_allclose(flmm.genetic_variance, 1.7303981309775553, rtol=1e-5)
    assert_allclose(flmm.environmental_variance, 1.2950028351268132, rtol=1e-5)


if __name__ == '__main__':
    __import__('pytest').main([__file__, '-s'])



from __future__ import division

from numpy import log
from numpy import pi
from numpy import var
from numpy.linalg import solve
from numpy.linalg import slogdet

from optimix import Composite


class SlowLMM(Composite):
    r"""General Linear Mixed Models.

    Models

    .. math::

        \mathbf y \sim \mathcal N\big(~ \mathbf m;~ \mathrm K ~\big)

    for any mean :math:`\mathbf m` and covariance matrix :math:`\mathrm K`.

    Args:
        y (array_like): real-valued outcome.
        mean (mean_function): mean function defined. (Refer to :doc:`mean`.)
        cov (covariance_function): covariance function defined. (Refer to :doc:`cov`.)
    """
    def __init__(self, y, mean, cov):
        if var(y) < 1e-8:
            raise ValueError("The phenotype variance is too low: %e." % var(y))

        super(SlowLMM, self).__init__(mean=mean, cov=cov)
        self._y = y
        self._cov = cov
        self._mean = mean

    def _lml_gradient_mean(self, mean, cov, gmean):

        Kiym = solve(cov, self._y - mean)

        g = []
        for i in range(len(gmean)):
            g.append(gmean[i].T.dot(Kiym))
        return g

    def _lml_gradient_cov(self, mean, cov, gcov):

        Kiym = solve(cov, self._y - mean)

        g = []
        for i in range(len(gcov)):
            g.append(-solve(cov, gcov[i]).diagonal().sum() + Kiym.dot(gcov[
                i].dot(Kiym)))
        return [gi / 2 for gi in g]

    def value(self, mean, cov):
        ym = self._y - mean
        Kiym = solve(cov, ym)

        (s, logdet) = slogdet(cov)
        assert s == 1.

        n = len(self._y)
        return -(logdet + ym.dot(Kiym) + n * log(2 * pi)) / 2

    def gradient(self, mean, cov, gmean, gcov):
        grad_cov = self._lml_gradient_cov(mean, cov, gcov)
        grad_mean = self._lml_gradient_mean(mean, cov, gmean)
        return grad_cov + grad_mean


# def variables(self):
#     v0 = self._mean.variables().select(fixed=False)
#     v1 = self._cov.variables().select(fixed=False)
#     return merge_variables(dict(mean=v0, cov=v1))

# def learn(self):
#     if len(self.variables()) == 0:
#         return
#     elif len(self.variables()) == 1:
#         maximize_scalar(self)
#     else:
#         maximize(self)

#     def predict(self):
#         mean = self._mean
#         cov = self._cov
#
#         m_p = mean.data('predict').value()
#         _Kim = self._Kim()
#         K_pp = cov.data('predict').value()
#
#         K_lp = cov.data('learn_predict').value()
#
#         emean = m_p + K_lp.T.dot(_Kim)
#         K = cov.feed().value()
#         ecov = K_pp - K_lp.T.dot(solve(K, K_lp))
#
#         return SlowLMMPredictor(emean, ecov)
#
#
# class SlowLMMPredictor(object):
#
#     def __init__(self, mean, cov):
#         self._mean = mean
#         self._cov = cov
#         self._mvn = multivariate_normal(mean, sum2diag(cov, epsilon.small))
#
#     @property
#     def mean(self):
#         return self._mean
#
#     @property
#     def cov(self):
#         return self._cov
#
#     def pdf(self, y):
#         return self._mvn.pdf(y)
#
#     def logpdf(self, y):
#         return self._mvn.logpdf(y)



from __future__ import division

from numpy import isscalar
from numpy import ix_
from numpy import dot
from numpy import ones
from numpy import zeros
from numpy import zeros_like
from numpy import tril_indices_from
from numpy import atleast_1d
from numpy import ascontiguousarray

from optimix import Vector
from optimix import Function

class FreeFormCov(Function):
    """
    General semi-definite positive matrix with no contraints.
    A free-form covariance matrix of dimension d has 1/2 * d * (d + 1) params
    """
    def __init__(self, size):
        """
        Args:
            dim:        dimension of the free-form covariance
            jitter:     extent of diagonal offset which is added for numerical stability
                        (default value: 1e-4)
        """
        tsize = ((size + 1) * size) / 2;
        tsize = int(tsize)
        self._L = zeros((size, size))
        self._tril = tril_indices_from(self._L)
        self._L[self._tril] = 1
        Function.__init__(self, Lu=Vector(ones(tsize)))

    @property
    def L(self):
        self._L[self._tril] = self.get('Lu')
        return self._L

    @L.setter
    def L(self, value):
        self._L[:] = value
        self.set('Lu', self._L[self._tril])

    @property
    def Lu(self):
        return self.get('Lu')

    @Lu.setter
    def Lu(self, value):
        self.set('Lu', value)

    def value(self, x0, x1):
        L = self.L
        K = dot(L, L.T)

        if isscalar(x0) and isscalar(x1):
            return K[x0, x1]

        one_scalar = isscalar(x0) or isscalar(x1)

        x0 = ascontiguousarray(atleast_1d(x0), int)
        x1 = ascontiguousarray(atleast_1d(x1), int)

        x0 = x0.ravel()
        x1 = x1.ravel()

        if one_scalar:
            return K[ix_(x0, x1)].ravel()

        return K[ix_(x0, x1)]

    def derivative_Lu(self, x0, x1):
        
        x0 = ascontiguousarray(atleast_1d(x0), int)
        x1 = ascontiguousarray(atleast_1d(x1), int)

        x0 = x0.ravel()
        x1 = x1.ravel()

        Lu = self.get('Lu')
        L = self.L
        size = L.shape[0]
        d = zeros((size, size, len(Lu)))
        for ii in range(len(self._tril[0])):
            row = self._tril[0][ii]
            d[:, row, ii] = L[:, row]
            d[row, :, ii] += d[:, row, ii]

        return d[ix_(x0, x1)]
